(* Nov 29, @Jinqi*)

open Type

type bstmt =
  Glb of int			(* Indicate the number of gloabal variables *)
| Pop of int			(* Pop out several elements from the stack *)
| Bin of Type.binop		(* Perform arithmetic on the top two elements of stack *)
| Uni of Type.unop		(* Perform unary operation on the top of stack *)
| Psi of int			(* Push an integer *)
| Psp of int * int		(* Push a pointer *)
| Psco of int * int * int	(* push a color *)
| Psk of string		(* push a key *)
| Psf of float		(* Push a floating number *)
| Psb of bool			(* Push a boolean number *)
| Psc of char			(* Push a char *)
| Pss of string		(* Push a string *)  
| Lod of int			(* Fetch global variable *)
| Str of int			(* Store global variable *)
| Lfp of int			(* Load frame pointer relative *)
| Sfp of int			(* Store frame pointer relative *)
| Jsr of int			(* Call function by absolute address *)
| Ent of int			(* Push FP, FP -> SP, SP += i *)
| Rts of int			(* Restore FP, SP, consume formals, push result *)
| Beq of int			(* Branch relative if top-of-stack is zero *)
| Bne of int			(* Branch relative if top-of-stack is non-zero *)
| Bra of int			(* Branch relative *)
| Hlt				(* Terminate *)

type prog = {
    num_globals : int;		(* Number of global variables *)
    text : bstmt array;	(* Code for all the functions *)
}

let string_of_stmt = function
  Glb(i) -> "glb " ^ string_of_int i
| Psi(i) -> "psi " ^ string_of_int i
| Psp(i,j) -> "psp " ^ string_of_int i ^ " " ^ string_of_int j
| Psco(i,j,k) -> "psco " ^ string_of_int i ^ " " ^ string_of_int j ^ " " ^ string_of_int k
| Psf(i) -> "psf " ^ string_of_float i
| Psb(i) -> if i then "psb " ^ "1" else "psb " ^ "0"
| Psc(i) -> "psc " ^ Char.escaped i
| Pss(i) -> "pss " ^ i ^ "\\0"
| Psk(k) -> "psk " ^ k ^ "\\0"
| Pop(i)-> "pop " ^ string_of_int i
| Bin(Type.Add) -> "add"
| Bin(Type.Sub) -> "sub"
| Bin(Type.Mul) -> "mul"
| Bin(Type.Div) -> "div"
| Bin(Type.Mod) -> "mod"
| Bin(Type.And) -> "and"
| Bin(Type.Or)  -> "or"
| Bin(Type.Eq) -> "eql"
| Bin(Type.Neq) -> "neq"
| Bin(Type.Lt) -> "lt"
| Bin(Type.Leq) -> "leq"
| Bin(Type.Gt) -> "gt"
| Bin(Type.Geq) -> "geq"
| Uni(Type.Neg) -> "neg"
| Uni(Type.Not) -> "not"
| Uni(Type.Atx) -> "atx"
| Uni(Type.Aty) -> "aty" 
| Uni(Type.Atr) -> "atr"
| Uni(Type.Atg) -> "atg" 
| Uni(Type.Atb) -> "atb"
| Uni(Type.Atk) -> "atk" 
| Lod(i) -> "lod " ^ string_of_int i
| Str(i) -> "str " ^ string_of_int i
| Lfp(i) -> "lfp " ^ string_of_int i
| Sfp(i) -> "sfp " ^ string_of_int i
| Jsr(i) -> "jsr " ^ string_of_int i
| Ent(i) -> "ent " ^ string_of_int i
| Rts(i) -> "rts " ^ string_of_int i
| Bne(i) -> "bne " ^ string_of_int i
| Beq(i) -> "beq " ^ string_of_int i
| Bra(i) -> "bra " ^ string_of_int i
| Hlt    -> "hlt"

let string_of_prog p =
	"" ^ let funca = Array.mapi (fun i s -> "" ^ string_of_stmt s) p.text
		in String.concat "\n" (Array.to_list funca)
