open Bytecode
open Type
open Sast
module StringMap =Map.Make(String)
(*test region*)
(*@Keqiu Hu & Lizhong Zhang*)
(*Referece: Microc--Stephen Edwards & TML--Yan Zou *)
(*Symbol table as MICROC*)
type env={
    function_index:int StringMap.t;
    global_index:int StringMap.t;
   (* local_index:int StringMap.t;*)
}

(*function to merge var_decl into a Block t*string list*)
(*let merge (t,str,_)=function*)


(*add names in list to the map (referto zou)*)
let add_to_map list map start_index =
    snd (List.fold_left (fun (i, map) name ->
    (*print_endline (name ^ ": " ^ (string_of_int i)); (* debug *) *) (i + 1, StringMap.add name i map)) (start_index, map) list)

(*function to perform preglobal expression handling*)
let add_to_map list map start_index =
snd (List.fold_left (fun (i, map) name ->
(*print_endline (name ^ ": " ^ (string_of_int i)); (* debug *) *) (i + 1, StringMap.add name i map)) (start_index, map) list)
(**)
let exprId =function
    (Id(a),b,c)->a
    |_->""
let takestr (a,b,c)=a
(*not needed anymore...
let expr2 env =function
    Assign(s,e)->[Str (StringMap.find s env.global_index)]*)
let rec enum stride n=function
    []->[]
    |hd::tl->(n,hd)::enum stride (n+stride) tl

(*add pairs into a map as MICORC*)
let string_map_pairs map pairs=
    List.fold_left (fun m (i,n) ->StringMap.add n i m) map pairs

(*main translate function, modified from MICROC*)
(*let p=function (a,b)->b

let res=function
    WithInit(a,b)->a
    |WithoutInit(c)->c

let initos=function
    (t,v)->List.map res v*)
(*take the t*string list out from a BLOCK*)
let takelistout=function
              Block(a,e)->e
              |d->[Int_T,"noinit"]
(*turn variable list into t*string list*)
let var_list t=
    List.map (fun(a,b,c)->(a,b)) t

let init_value t=function
    	| Some(e) -> e
	| None ->
			let lit = match t with
                                
                                Void_T->Integer(0)
                                | Point_T->Point(0,0)   
                                | Int_T -> Integer(0)
                                |Key_T->String("Null")
                                |Color_T->Color(0,0,0)
				| Float_T -> Float(0.0)
				| Char_T -> Char('\000')
				| String_T -> String("\0")
				| Boolean_T -> Boolean(false)
				
			in Literal(lit), t,true

(*translate the program passed in*)

let translate (globals,functions)=
   (* let glbexpr=(List.map (fun (a,b,c)->c) globals) in*)
    let precute=List.map (fun (a,b,c)->b) globals in
  (* Allocate "addresses" for each global variable *)
    let global_indexes = string_map_pairs StringMap.empty (enum 1 0 precute) in

  (* Assign indexes to function names; built-in "print" is special *)
(*  let built_in_functions = StringMap.add "print" (-1) StringMap.empty in*)
    (*buildin function manipulation*)
    let built_in_functions =
        let builtin =
            ["__leftClickPoint";"__rightClickPoint";"__doubleClick";"__movePoint";"__leftDown";"__rightDown";"__leftUp";"__rightUp";"__keyDown";"__keyUp";"__keyStroke";"__comboStroke";"__getPoint";"__getcolor";"__print";"__stringOfInt";"__stringOfFloat";"__stringOfChar";"__stringOfBoolean";"__stringOfPoint";"__stringOfKey";"__stringOfColor";"__getASCII";"__getChar";"__getStringLength";"__getCharOfString";"__setCharOfString";"__subOfString";"__delay";"__compareColor"] in string_map_pairs StringMap.empty (enum (-1) (-1) builtin) in

  let function_indexes = string_map_pairs built_in_functions
      (enum 1 0 (List.map (fun f -> f.fname) functions)) in

  (* Translate a function in AST form into a list of bytecode statements *)
  let rec block local_index next_index num_params sl=
        let rec expr (exprdetail,t,int)=match exprdetail with
            Literal i->begin match i with
                Integer(j)->[Psi j]
                |Char(c)->[Psc c]
                |String(s)->[Pss s]
                |Float(f)->[Psf f]
                |Boolean(b)->[Psb b]
                |Point(x,y)->[Psp(x,y)]
                |Key(k)->[Pss k]
                |Color(r,g,l)->[Psco (r,g,l)](*later chage*)
                end
            |Id(s)->
                (try [Lfp (StringMap.find s local_index)]
                with Not_found -> try [Lod (StringMap.find s global_indexes)]
                with Not_found -> raise (Failure ("undeclared variable"^s)))
    
            |Binop(e1,op,e2) -> expr e1 @ expr e2 @[Bin op]

	    |Unop (op, e) -> (expr e) @ [Uni op] 

            |Assign (s,e)->expr e @
                (try [Sfp (StringMap.find s local_index)]
                with Not_found->try [Str (StringMap.find s
                global_indexes)]
                with Not_found->raise (Failure("Undeclared variable"^(s))))
            |Call(fname,actuals)->
                    if(fname="__leftClickPoint") then 
                        if(List.length actuals>0) then
                            (List.concat (List.map (fun e->
                                (expr e) @ [Jsr (-1);Pop 1]) actuals))@
                            [Psi 0]
                        else []
                    else if(fname="__rightClickPoint") then
                       List.concat (List.map (fun e->(expr e)) actuals) @[Jsr (-2);Pop 1]@[Psi 0]

                    else if(fname="__doubleClick") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-3);Pop 1]@[Psi 0]                   
                    else if(fname="__movePoint") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-4)]@[Psi 0]
                    else if(fname="__leftDown") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-5)]@[Psi 0]                   
                    else if(fname="__rightDown") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-6)]@[Psi 0]                   
                    else if(fname="__leftUp") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-7);Pop 1;Pop 1]@[Psi 0]                   
                    else if(fname="__rightUp") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-8);Pop 1;Pop 1]@[Psi 0]                  
                    else if(fname="__keyDown") then
                       List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                       (-9);Pop 1;Pop 1]@[Psi 0]                  
                    else if(fname="__keyUp") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-10);Pop 1;Pop 1]@[Psi 0]                 
                    else if(fname="__keyStroke") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-11);Pop 1]@[Psi 0]                  
                    else if(fname="__comboStroke") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-12);Pop 1;Pop 1]@[Psi 0]                
                    else if(fname="__getPoint") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-13)]           
               
                    else if(fname="__getcolor") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-14)]                  
                    else if(fname="__print") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-15);Pop 1]@[Psi 0]                 
                    else if(fname="__stringOfInt") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-16);]               
                    else if(fname="__stringOfFloat") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-17);Pop 1;Pop 1]@[Psi 0]               
                    else if(fname="__stringOfChar") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-18);Pop 1;Pop 1]@[Psi 0]                 
                    else if(fname="__stringOfBoolean") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-19);Pop 1;Pop 1]@[Psi 0]                
                    else if(fname="__stringOfPoint") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-20);Pop 1;Pop 1]@[Psi 0]               
                    else if(fname="__stringOfKey") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-21);Pop 1;Pop 1]@[Psi 0]                  
                    else if(fname="__stringOfColor") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-22);Pop 1;Pop 1]@[Psi 0]                 
                    else if(fname="__getASCII") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-23);Pop 1;Pop 1]@[Psi 0]                 
                    else if(fname="__getChar") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-24);Pop 1;Pop 1]@[Psi 0]                   
                    else if(fname="__getStringLength") then
                       List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                       (-25);Pop 1;Pop 1]@[Psi 0]                 
                    else if(fname="__getCharOfString") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-26);Pop 1;Pop 1]@[Psi 0]                  
                    else if(fname="__setCharOfString") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-27);Pop 1;Pop 1]@[Psi 0]                
                    else if(fname="__subOfString") then
                        List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                        (-28);Pop 1;Pop 1]@[Psi 0]                   
                    else if(fname="__delay") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-29);Pop 1]@[Psi 0]
                    else if(fname="__compareColor") then
                         List.concat (List.map (fun e->(expr e)) actuals) @[Jsr
                         (-30);Pop 1;Pop 1]@[Psi 0]
                    else
                            (try
                (List.concat (List.map expr (List.rev actuals)))@[Jsr
                (StringMap.find fname function_indexes)]
                with Not_found->raise (Failure("undefined function "^fname)))
            |NoExpr->[]
             in 
             (*modified in order to create new variables inside a block(trial)*)
             let rec stmt = function
        	Block (sl,vl)     ->  
                    let vname_list=List.map (fun (_,n)->n) vl in
                    let new_local_index=
                        add_to_map vname_list local_index next_index
                    in
                    let num_locals=List.length vl in
                    (block new_local_index (next_index+num_locals) num_params
                    sl) @(if num_locals>0 then
                        [Pop num_locals]
                    else[])

                
              | Expr e       -> expr e @ [Pop 1]
              | Return e     -> expr e @ [Rts num_params]
        (*no idea how to change this if*)
              | If (p, t) -> begin match (p,t) with
                |([],[])->[Bra 1]
                |([],t)->let t'=(List.hd t) in stmt t'
                |(p,t)->let p'=List.hd p and t'= stmt (List.hd t) and f'= stmt (If(List.tl p,List.tl t)) in expr p' @ [Beq(2 + List.length t')] @t' @ [Bra(1 + List.length f')] @ f'
                end
              | VarDecl(t,name,init)->expr(init_value t init)(*begin match e1 with (a,b,Some(c))->expr c
              |(a,b,None)-> [Bra 1] end*)
              | For (e1, e2, e3, b) ->
                      stmt (Block([Expr(e1); While(e2,
                      Block([b;Expr(e3)],takelistout b))],takelistout b))

              (*Just need to add the initialization in v1 into the variable
               * stack of b and others are the same*)
(*              (takelistout b)@(var_list v1) *)
              |  ForWithDecl (v1, e2, e3, b) ->
                      let li=(takelistout
                      b)@(var_list v1) in
                      stmt (Block([While(e2,Block([b;Expr(e3)],li))],li))
	
              |  Switch (e, el, sl) -> begin match (e, el, sl) with
				  (_, [], []) -> [Bra 1]
				| (_, [], t) -> let t'=(List.hd t) in stmt t'
				| (e, el, sl) -> let p = List.hd el and t' = stmt (List.hd sl) and f' = stmt (Switch(e, List.tl el, List.tl sl)) in expr e @ expr p @ [Bin(Type.Eq)] @ [Bne(2)] @ [Bra(2 + List.length t')] @ t' @[Bra(1+List.length f')] @ f'
					end


              |  Break (i) -> [Pop (i+1)]

              |  Continue (i) -> [Pop (i+1)]
		
	    
              | While (e, b) ->
        	let b' = stmt b and e' = expr e in
	          [Bra (1+ List.length b')] @ b' @ e' @
        	  [Bne (-(List.length b' + List.length e'))]

        in 
        List.concat (List.map stmt sl)
    in
    let data=
        let assign_globals=
            List.map (fun (t,name,init)->
                Expr(Assign(name,init_value t init),t,true)) globals
        in
        [Glb (List.length assign_globals)]@
        block StringMap.empty 0 0 assign_globals @
        try 
            [Jsr (StringMap.find "main" function_indexes);Hlt]
        with Not_found->
            raise (Failure "main function not found")
    in
    let text=
        let trans_func f=
            let param_names=List.map snd f.formals in
            let num_params=List.length param_names in
            let param_index=
                add_to_map param_names StringMap.empty (-1 -num_params)
            in
            let local_names=List.map snd f.locals in
            let num_locals=List.length local_names in
            let local_index=
                add_to_map local_names param_index 1
            in
            let bodycodes=
                block local_index (num_locals +1) num_params f.body
            in
            List.iter (function b->()) bodycodes;(*not quite understand*)
            [Ent num_locals]@bodycodes@
            (* 
                                Void_T->Integer(0)
                                | Point_T->Point(0,0)   
                                | Int_T -> Integer(0)
                                |Key_T->String("")
                                |Color_T->Color(0,0,0)
				| Float_T -> Float(0.0)
				| Char_T -> Char('\000')
				| String_T -> String("")
				| Boolean_T -> Boolean(false)*)
            (match f.rtype with

            |Int_T->[Psi 0]
            |Key_T->[Psk ""]
            |Color_T->[Psco (0,0,0)]
            |Point_T->[Psp (0,0)]
            |Float_T -> [Psf 0.0]
            | Char_T -> [Psc '\000']
    	    | String_T -> [Pss ""]
       	    | Boolean_T -> [Psb false]
    	    | Void_T -> [Psi 0]) @
        [Rts num_params]
        in    
        let func_bodies=
            (List.map trans_func functions)
        in
        let (_,func_offsets)=
            List.fold_left (fun (i,offsets) body->
                ((List.length body) +i,i::offsets))
            ((List.length data),[])func_bodies
        in
        let array_offsets=Array.of_list (List.rev func_offsets) in
        List.map (function 
            |Jsr i when i>=0->Jsr array_offsets.(i)
            |_ as s->s)
        (List.concat (data::func_bodies))
    in
    Bytecode.string_of_prog {num_globals=10;text=Array.of_list text}
